import { useOffsetPagination, UseOffsetPaginationOptions, UseOffsetPaginationReturn } from '@vueuse/core'
import { nextTick, reactive, ref, Ref, UnwrapRef, WatchOptions, watch } from 'vue'
import { DebouncedFunc } from 'lodash'
import debounce from 'lodash/debounce'
import { usePromise } from '../usePromise'

export type UseServerPaginationResolve = UnwrapRef<
  Pick<UseOffsetPaginationReturn, 'currentPage' | 'currentPageSize'> & { total: number }
>
export type UseServerPagination = UnwrapRef<Omit<UseOffsetPaginationReturn, 'next' | 'prev'> & { total: Ref<number> }>

export interface UseServerPaginationOptions<T> extends UseOffsetPaginationOptions, WatchOptions {
  resolve: (pagination: UseServerPaginationResolve) => T | Promise<T>
  /** 监听源, 当源数据改变触发 reset, 默认监听 pagination */
  sources?: any[]
}
export interface UseServerPaginationReturn<T> {
  state: Ref<T>
  loading: Ref<boolean>
  error: Ref<any>
  pagination: UseServerPagination
  next: () => void
  prev: () => void
  execute: DebouncedFunc<() => Promise<void>>
}

export const useServerPagination = <T extends any[]>(
  options: UseServerPaginationOptions<T>
): UseServerPaginationReturn<T> => {
  const total = ref(Infinity)
  const pagination = useOffsetPagination({ total, ...options }) as UseOffsetPaginationReturn
  const resolveOptions = reactive({
    currentPage: pagination.currentPage,
    currentPageSize: pagination.currentPageSize,
    total
  })

  const paginationRef = reactive({
    currentPage: pagination.currentPage,
    currentPageSize: pagination.currentPageSize,
    pageCount: pagination.pageCount,
    isFirstPage: pagination.isFirstPage,
    isLastPage: pagination.isLastPage,
    total
  })

  const state = ref<any>([]) as Ref<T>
  const {
    execute: _execute,
    loading,
    error
  } = usePromise(async () => {
    const resolved = await options.resolve(resolveOptions)
    return resolved || []
  })

  const execute = debounce(async () => {
    state.value = await _execute()
  }, 100)

  nextTick(() => {
    watch([pagination.currentPageSize, ...(options.sources || [])], execute, { immediate: true, ...options })
    watch(pagination.currentPage, (newValue, oldValue) => newValue !== oldValue && execute())
  })

  return {
    state,
    loading,
    error,
    execute,
    pagination: paginationRef,
    next: pagination.next,
    prev: pagination.prev
  }
}
